# Suomilog
# Copyright (C) 2018 Iikka Hauhio
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

#"Tulosta" [
#	teksti<-$nimi:nimento
#]
#
#ihminen :: OLENTO
#
#OLENTO:omanto [
#	omistaja->$nimi:*
#	omistaja->$pituus:*
#	omistaja->@lapset:*
#	omistaja->LAPSI:*
#]
#
#$nimi:* [
#	koko<-KIRJAINKOKO:ulko-olento
#	.
#]
#
#KIRJAINKOKO:x {
#	@'iso:x [
#		koko->@kirjain:x
#	]
#	@'pieni:x [
#		koko->@kirjain:x
#	]
#}
#
#lapsi :: OLENTO
#
#LAPSI:x {
#	$ensimmäinen:x [
#		mones->$lapsi:x
#	]
#	$toinen:x [
#		mones->$lapsi:x
#	]
#}

#PYTHONILLA:

class TokenList:
	def __init__(self, tokens):
		self.tokens = tokens
	def __getitem__(self, i):
		return self.tokens[i]
	def __len__(self):
		return len(self.tokens)

class Word:
	def __init__(self, token, baseform, form, number, relations={}):
		self.token = token
		self.baseform = baseform
		self.form = form
		self.number = number
		self.relations = relations
	def update(self, relation, word):
		return Word(self.token, self.baseform, self.form, self.number, {**self.relations, relation: word})
	def __repr__(self):
		return self.baseform + ":" + self.form + ":" + self.number + repr(self.relations)

class Alternative:
	def __init__(self, pattern, relation, dominative):
		self.pattern = pattern
		self.relation = relation
		self.dominative = dominative
	def __repr__(self):
		return self.relation + ("->" if self.dominative else "<-") + repr(self.pattern)

class Namespace:
	def __init__(self, rules, parameters):
		self._rules = rules
		self._parameters = parameters
	def rules(self):
		return self._rules
	def parameters(self):
		return self._parameters

class NamespaceReference:
	def __init__(self, name):
		self.ref_name = name
	def rules(self):
		return NAMESPACES[self.ref_name]._rules
	def parameters(self):
		return NAMESPACES[self.ref_name]._parameters
	def __repr__(self):
		return self.ref_name

class Pattern:
	def __init__(self, baseform, form, number):
		self.baseform = baseform
		self.form = form
		self.number = number
	def isSubnamespacePattern(self):
		return isinstance(self.baseform, NamespaceReference) or isinstance(self.baseform, Namespace)
	def match(self, word):
		if ((isinstance(self.baseform, str) and self.baseform == word.baseform or word.baseform in self.baseform)
			and (len(self.form) <= 1 or self.form == word.form)
			and (len(self.number) <= 1 or self.number == word.number)):
			subs = {}
			if len(self.form) == 1:
				subs[self.form] = word.form
			if len(self.number) == 1:
				subs[self.number] = word.number
			return subs
		else:
			return None
	def subs(self, subs):
		return Pattern(
			self.baseform,
			self.form if len(self.form) > 1 else subs.get(self.form, self.form),
			self.number if len(self.number) > 1 else subs.get(self.number, self.number))
	def __repr__(self):
		return repr(self.baseform) + ":" + self.form + ":" + self.number

class LiteralPattern:
	def __init__(self, baseform):
		self.baseform = baseform
	def isSubnamespacePattern(self):
		return False
	def match(self, word):
		return {} if word.token == self.baseform else None
	def subs(self, subs):
		return self

class Environment:
	def __init__(self, rules, tokens, place=0, current_object=None, substitutions={}, onEnd=lambda env: []):
		if current_object is None:
			current_object = tokens[place]
			place += 1
		self.rules = rules
		self.tokens = tokens
		self.place = place
		self.current_object = current_object
		self.substitutions = substitutions
		self.onEnd = onEnd
	def hasNext(self):
		return len(self.tokens) > self.place
	def nextWord(self):
		return self.tokens[self.place]
	def next(self, word, alt, altpattern):
		def update(env, next_word, delta):
			if alt.dominative:
				new_obj = next_word.update(alt.relation, self.current_object)
			else:
				new_obj = self.current_object.update(alt.relation, next_word)
			return Environment(self.rules, self.tokens, env.place+delta, new_obj, substitutions=self.substitutions, onEnd=self.onEnd)
		if altpattern.isSubnamespacePattern():
			form_var, number_var = altpattern.baseform.parameters()
			subs = {form_var: altpattern.form, number_var: altpattern.number}
			return Environment(alt.pattern.baseform.rules(), self.tokens, self.place, substitutions=subs, onEnd=lambda env: [update(env, env.current_object, 0)])
		else:
			return update(self, word, 1)

class Rule:
	def __init__(self, pattern, alternatives, can_end):
		self.pattern = pattern
		self.alternatives = alternatives
		self.can_end = can_end
	def match(self, env):
		pattern = self.pattern.subs(env.substitutions)
		subs = pattern.match(env.current_object)
		#print(subs, env.place)
		if subs is not None:
			ans = env.onEnd(env) if self.can_end else []
			if env.hasNext():
				word = env.nextWord()
				for alt in self.alternatives:
					altpattern = alt.pattern.subs(subs).subs(env.substitutions)
					#print(word, altpattern)
					if altpattern.isSubnamespacePattern() or altpattern.match(word) is not None:
						ans.append(env.next(word, alt, altpattern))
			return True, ans
			
		else:
			return False, []
	def __repr__(self):
		return repr(self.pattern) + " [" + ",".join(map(repr, self.alternatives)) + "]"

def parse(rules, tokens):
	ans = []
	def onEnd(env):
		nonlocal ans
		if not env.hasNext():
			ans.append(env.current_object)
		return []
	envs = [Environment(rules, tokens, onEnd=onEnd)]
	while envs:
		new_envs = []
		for env in envs:
			found = False
			for rule in env.rules:
				print(env.current_object, rule)
				matches, news = rule.match(env)
				print("->", news)
				if matches:
					new_envs += news
					found = True
			if not found:
				new_envs += env.onEnd(env)
		envs = new_envs
	return ans

def main():
	global NAMESPACES
	NAMESPACES = {}
	NAMESPACES["LAPSI"] = Namespace([
		Rule(Pattern("ensimmäinen", "f", "x"), [Alternative(Pattern("lapsi", "f", "x"), "mones", True)], False),
		Rule(Pattern("toinen", "f", "x"), [Alternative(Pattern("lapsi", "f", "x"), "mones", True)], False)
	], ("f", "n"))
	NAMESPACES["KIRJAINKOKO"] = Namespace([
		Rule(Pattern("iso", "f", "monikko"), [Alternative(Pattern("kirjain", "f", "monikko"), "koko", True)], False),
		Rule(Pattern("pieni", "f", "monikko"), [Alternative(Pattern("kirjain", "f", "monikko"), "koko", True)], False)
	], ("f", "n"))
	olento = ["pelaaja", "lapsi"]
	merkkijono = ["nimi"]
	rules = [
		Rule(Pattern(olento, "omanto", "yksikkö"), [
			Alternative(Pattern(NamespaceReference("LAPSI"), "", ""), "omistaja", True),
			Alternative(Pattern("nimi", "", "yksikkö"), "omistaja", True)
		], False),
		Rule(Pattern(merkkijono, "", "yksikkö"), [
			Alternative(Pattern(NamespaceReference("KIRJAINKOKO"), "ulkoolento", ""), "tyyli", False)
		], True)
	]
	tokens = TokenList([
		Word("pelaajan", "pelaaja", "omanto", "yksikkö"),
		Word("toisen", "toinen", "omanto", "yksikkö"),
		Word("lapsen", "lapsi", "omanto", "yksikkö"),
		Word("nimi", "nimi", "nimento", "yksikkö"),
		Word("pienillä", "pieni", "ulkoolento", "monikko"),
		Word("kirjaimilla", "kirjain", "ulkoolento", "monikko")
	])
	print("ans", parse(rules, tokens))

main()



# Suomilog
# Copyright (C) 2019 Iikka Hauhio
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import re
from voikko import libvoikko as lv
from voikko.inflect_word import inflect_word
from . import patternparser as pp

voikko = lv.Voikko("fi-x-morpho")

def tokenize(text):
	tokens = []
	for token in voikko.tokens(text):
		if token.tokenType == lv.Token.WHITESPACE:
			continue
		if "-" in token.tokenText:
			index = token.tokenText.rindex("-")+1
			lastPart = token.tokenText[index:]
			baseformPrefix = token.tokenText[:index].lower()
		else:
			lastPart = token.tokenText
			baseformPrefix = ""
		alternatives = []
		for word in voikko.analyze(token.tokenText):
			if "BASEFORM" in word:
				alternatives.append(baseformAndBits(word))
		# Jos jäsennys epäonnistui, koetetaan jäsennystä vain viimeisen palan kautta
		if len(alternatives) == 0 and baseformPrefix:
			for word in voikko.analyze(lastPart):
				if "BASEFORM" in word:
					alternatives.append(baseformAndBits(word, baseformPrefix))
		tokens.append(pp.Token(token.tokenText, alternatives))
	return tokens

def baseformAndBits(word, baseformPrefix=None):
	bits = set()
	addBits(word, bits, "NUMBER", {"singular": "yksikkö", "plural": "monikko"})
	addBits(word, bits, "SIJAMUOTO")
	addBits(word, bits, "CLASS")
	addBits(word, bits, "PARTICIPLE")
	addBits(word, bits, "PERSON")
	addBits(word, bits, "MOOD", {
		"MINEN-infinitive": "-minen",
		"MA-infinitive": "-ma",
		"E-infinitive": "-e",
		"A-infinitive": "-a",
		"imperative": "imperatiivi",
		"indicative": "indikatiivi",
		"conditional": "konditionaali",
		"potential": "potentiaali"
	})
	if word["CLASS"] == "lukusana" and ("SIJAMUOTO" not in word or not word["SIJAMUOTO"]):
		bits.add("nimento")
	if not baseformPrefix:
		return word["BASEFORM"].lower(), bits
	else:
		return baseformPrefix + word["BASEFORM"].lower(), bits

def addBits(word, bits, name, table=None):
	if name in word:
		if table:
			if word[name] in table:
				bits.add(table[word[name]])
		else:
			bits.add(word[name])

CASES = [
	"nimento",
	"omanto",
	"osanto",
	"olento",
	"tulento",
	"ulkoolento",
	"ulkotulento",
	"ulkoeronto",
	"sisaolento",
	"sisatulento",
	"sisaeronto",
	"vajanto"
]

CASES_LATIN = {
	"nimento": "nominatiivi",
	"omanto": "genetiivi",
	"osanto": "partitiivi",
	"olento": "essiivi",
	"tulento": "translatiivi",
	"ulkotulento": "allatiivi",
	"ulkoolento": "adessiivi",
	"ulkoeronto": "ablatiivi",
	"sisatulento": "illatiivi",
	"sisaolento": "inessiivi",
	"sisaeronto": "elatiivi",
	"vajanto": "abessiivi",
	"keinonto": "instruktiivi",
	"seuranto": "komitatiivi",
	"kerrontosti": "adverbi"
}

CASES_ENGLISH = {
	"nimento": "nominative",
	"omanto": "genitive",
	"osanto": "partitive",
	"olento": "essive",
	"tulento": "translative",
	"ulkotulento": "allative",
	"ulkoolento": "adessive",
	"ulkoeronto": "ablative",
	"sisatulento": "illative",
	"sisaolento": "inessive",
	"sisaeronto": "elative",
	"vajanto": "abessive",
	"keinonto": "instructive",
	"seuranto": "comitative",
	"kerrontosti": "adverb"
}

CASES_A = {
	"nimento": "",
	"omanto": ":n",
	"osanto": ":ta",
	"olento": ":na",
	"tulento": ":ksi",
	"ulkotulento": ":lle",
	"ulkoolento": ":lla",
	"ulkoeronto": ":lta",
	"sisatulento": ":han",
	"sisaolento": ":ssa",
	"sisaeronto": ":sta",
	"vajanto": ":tta",
	"keinonto": ":in",
	"seuranto": ":ineen",
	"kerrontosti": ":sti"
}

CASES_F = {
	"nimento": "",
	"omanto": ":n",
	"osanto": ":ää",
	"olento": ":nä",
	"tulento": ":ksi",
	"ulkotulento": ":lle",
	"ulkoolento": ":llä",
	"ulkoeronto": ":ltä",
	"sisatulento": ":ään",
	"sisaolento": ":ssä",
	"sisaeronto": ":stä",
	"vajanto": ":ttä",
	"keinonto": ":in",
	"seuranto": ":ineen",
	"kerrontosti": ":sti"
}

CASES_ELLIPSI = {
	"nimento": "",
	"omanto": ":n",
	"osanto": ":ä",
	"olento": ":nä",
	"tulento": ":ksi",
	"ulkotulento": ":lle",
	"ulkoolento": ":llä",
	"ulkoeronto": ":ltä",
	"sisatulento": ":iin",
	"sisaolento": ":ssä",
	"sisaeronto": ":stä",
	"vajanto": ":ttä",
	"keinonto": ":ein",
	"seuranto": ":eineen",
	"kerrontosti": ":sti"
}

CASE_REGEXES = {
	"singular": {
		"omanto": r"[^:]+:n",
		"osanto": r"[^:]+:(aa?|ää?|t[aä])",
		"olento": r"[^:]+:(n[aä])",
		"tulento": r"[^:]+:ksi",
		"ulkotulento": r"[^:]+:lle",
		"ulkoolento": r"[^:]+:ll[aä]",
		"ulkoeronto": r"[^:]+:lt[aä]",
		"sisatulento": r"[^:]+:(aan|ään|h[aeiouyäöå]n)",
		"sisaolento": r"[^:]+:ss[aä]",
		"sisaeronto": r"[^:]+:st[aä]",
		"vajanto": r"[^:]+:tt[aä]"
	},
	"plural": {
		"omanto": r"[^:]+:ien",
		"osanto": r"[^:]+:(ia?|iä?|it[aä])",
		"olento": r"[^:]+:(in[aä])",
		"tulento": r"[^:]+:iksi",
		"ulkotulento": r"[^:]+:ille",
		"ulkoolento": r"[^:]+:ill[aä]",
		"ulkoeronto": r"[^:]+:ilt[aä]",
		"sisatulento": r"[^:]+:(iin|ih[aeiouyäöå]n)",
		"sisaolento": r"[^:]+:iss[aä]",
		"sisaeronto": r"[^:]+:ist[aä]",
		"vajanto": r"[^:]+:itt[aä]",
		"keinonto": r"[^:]+:in",
		"seuranto": r"[^:]+:ine[^:]*"
	},
	"": {
		"kerrontosti": "[^:]+:sti"
	}
}

ORDINAL_CASE_REGEXES = {
	"nimento": r"[^:]+:s",
	"omanto": r"[^:]+:nnen",
	"osanto": r"[^:]+:tt[aä]",
	"tulento": r"[^:]+:nneksi",
	"ulkotulento": r"[^:]+:nnelle",
	"ulkoolento": r"[^:]+:nnell[aä]",
	"ulkoeronto": r"[^:]+:nnelt[aä]",
	"sisatulento": r"[^:]+:nteen",
	"sisaolento": r"[^:]+:nness[aä]",
	"sisaeronto": r"[^:]+:nnest[aä]",
	"vajanto": r"[^:]+:nnett[aä]",
	"kerrontosti": r"[^:]+:nnesti"
}

def inflect(word, case, plural):
	case_latin = CASES_LATIN[case]
	if plural:
		case_latin += "_mon"
	
	if re.fullmatch(r"[0-9]+", word):
		if case == "sisatulento":
			if word[-1] in "123560":
				return word + ":een"
			elif word[-1] in "479":
				return word + ":ään"
			else: # 8
				return word + ":aan"
		elif word[-1] in "14579":
			return word + CASES_A[case].replace("a", "ä")
		else:
			return word + CASES_A[case]
	elif len(word) == 1:
		if word in "flmnrsx":
			return word + CASES_F[case]
		elif case == "sisatulento":
			if word in "aeiouyäöå":
				return word + ":h" + word + "n"
			elif word in "bcdgptvw":
				return word + ":hen"
			elif word in "hk":
				return word + ":hon"
			elif word == "j":
				return "j:hin"
			elif word == "q":
				return "q:hun"
			elif word == "z":
				return "z:aan"
		elif word in "ahkoquzå":
			return word + CASES_A[case]
		else:
			return word + CASES_A[case].replace("a", "ä")
	else:
		inflections = inflect_word(word)
		if case_latin not in inflections:
			return word + ":" + case
		return inflections[case_latin]



from setuptools import setup

setup(
	name='suomilog',
	version='0.8',
	description='Collection of natural language parsing utilities',
	url='http://github.com/fergusq/suomilog',
	author='Iikka Hauhio',
	author_email='iikka.hauhio@gmail.com',
	license='GPL',
	classifiers=[
		'Programming Language :: Python :: 3'
	],
	packages=['suomilog'],
	python_requires='>=3',
	install_requires=['voikko'],
)



# Suomilog
# Copyright (C) 2018 Iikka Hauhio
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import sys
import re
import itertools
from copy import deepcopy
from collections import defaultdict

class Token:
	def __init__(self, token, alternatives):
		self.token = token
		self.alternatives = alternatives
	def __repr__(self):
		return "Token(" + repr(self.token) + ", " + repr(self.alternatives) + ")"
	def __str__(self):
		return self.token + "[" + "/".join([baseform + "{" + ", ".join(bits) + "}" for baseform, bits in self.alternatives]) + "]"
	def toCode(self):
		return self.token or "/".join([baseform + "{" + ",".join(bits) + "}" for baseform, bits in self.alternatives])
	def containsMatch(self, alternatives):
		return any([any([tbf == baseform and tbits >= bits for tbf, tbits in self.alternatives]) for baseform, bits in alternatives])
	def baseform(self, *tbits):
		tbits = set(tbits)
		for baseform, bits in self.alternatives:
			if tbits <= bits:
				return baseform
		return None
	def bits(self):
		bits = set()
		for _, b in self.alternatives:
			bits.update(b)
		return bits

class PatternRef:
	def __init__(self, name, bits):
		self.name = name
		self.bits = bits
	def __repr__(self):
		return "PatternRef(" + repr(self.name) + ", " + repr(self.bits) + ")"
	def toCode(self):
		return "." + self.name + ("{" + ",".join(self.bits) + "}" if self.bits else "")

class MultiOutput:
	def __init__(self, outputs):
		self.outputs = outputs
	def __repr__(self):
		return "MultiOutput(" + repr(self.outputs) + ")"
	def eval(self, args):
		ans = []
		for i, output in enumerate(self.outputs):
			ans.append(output.eval([arg[i] for arg in args]))
		return ans

class StringOutput:
	def __init__(self, string):
		self.string = string
	def __repr__(self):
		return "StringOutput(" + repr(self.string) + ")"
	def eval(self, args):
		ans = self.string
		for i, a in enumerate(args):
			var = "$"+str(i+1)
			if var not in ans:
				ans = ans.replace("$*", ",".join(args[i:]))
				break
			ans = ans.replace(var, a)
		return ans

class Grammar:
	def __init__(self, patterns=None, names=None):
		self.patterns = patterns or {}
		self.names = names or {}
	def print(self):
		for category in sorted(self.patterns):
			print(category, "::=")
			for pattern in self.patterns[category]:
				print(" " + pattern.toCode())
	def copy(self):
		return Grammar({name: self.patterns[name].copy() for name in self.patterns}, self.names.copy())
	def update(self, grammar):
		for category in grammar.patterns:
			if category in self.patterns:
				self.patterns[category] += grammar.patterns[category]
			else:
				self.patterns[category] = grammar.patterns[category].copy()
		self.names.update(grammar.names)
	def matchAll(self, tokens, category, bits):
		if category not in self.patterns:
			return []
		
		ans = []
		for pattern in self.patterns[category]:
			ans += pattern.match(self, tokens, bits)
		
		return ans
	def allowsEmptyContent(self, category):
		if category not in self.patterns:
			return False
		return any([pattern.allowsEmptyContent() for pattern in self.patterns[category]])
	def addCategoryName(self, category, name):
		self.names[category] = name
	def refToString(self, ref):
		if ref.name in self.names:
			name = self.names[ref.name]
		else:
			name = ref.name
		if ref.bits:
			name += " (" + ",".join(ref.bits) + ")"
		return name
	def parseGrammarLine(self, line, *outputs):
		if debug_level >= 1:
			print(line)
		tokens = line.replace("\t", " ").split(" ")
		if len(tokens) > 2 and tokens[1] == "::=" and (outputs or "->" in tokens):
			end = tokens.index("->") if "->" in tokens else len(tokens)
			category = tokens[0][1:]
			words = []
			for token in tokens[2:end]:
				if "{" in token and token[-1] == "}":
					bits = set(token[token.index("{")+1:-1].split(","))
					token = token[:token.index("{")]
				else:
					bits = set()
				if token == "":
					continue
				elif token == ".":
					words.append(Token(".", []))
				elif token[0] == ".":
					words.append(PatternRef(token[1:], bits))
				elif bits:
					words.append(Token("", [(token, bits)]))
				else:
					words.append(Token(token, []))
			if category not in self.patterns:
				self.patterns[category] = []
			if "->" in tokens:
				outputs = outputs + (StringOutput(" ".join(tokens[end+1:])),)
			pattern = Pattern(category, words, MultiOutput(outputs) if len(outputs) > 1 else outputs[0])
			self.patterns[category].append(pattern)
			return pattern
		else:
			raise Exception("Syntax error on line `" + line + "'")

debug_level = 0
indent = 0

def setDebug(n):
	global debug_level
	debug_level = n

def groupCauses(causes):
	a = []
	groups = defaultdict(list)
	for cause in causes:
		if cause.is_part and not cause.causesContainPartErrors():
			groups[(id(cause.grammar), cause.place_string)].append(cause)
		else:
			a.append(cause)
	
	for group in groups.values():
		refs = []
		for cause in group:
			refs += cause.refs
		a.append(ParsingError(group[0].grammar, refs, group[0].place_string, True, []))
	
	return a

class ParsingError:
	def __init__(self, grammar, refs, place_string, is_part, causes):
		self.grammar = grammar
		self.refs = refs
		self.place_string = place_string
		self.is_part = is_part
		self.causes = groupCauses(causes)
	def print(self, finnish=False, file=sys.stderr):
		global indent
		ref_str = (" tai " if finnish else " or ").join(map(self.grammar.refToString, self.refs))
		if self.is_part:
			string = ("\n"+self.place_string).replace("\n", "\n"+"  "*indent)[1:]
			if finnish:
				print(string + "tämän pitäisi olla " + ref_str, file=file)
			else:
				print(string + "expected " + ref_str, file=file)
		if self.causesContainPartErrors():
			if finnish:
				print("  "*indent+"lauseke ei voi olla "+ref_str+" koska:", file=file)
			else:
				print("  "*indent+ref_str+" does not match because:", file=file)
			indent += 1
			causes = list(filter(ParsingError.containsPartErrors, self.causes))
			for cause in causes[:5]:
				cause.print(finnish=finnish)
				print(file=file)
			if len(causes) > 5:
				if finnish:
					print("  "*indent+"+ " + str(len(self.causes)-5) + " muuta mahdollista virhettä", file=file)
				else:
					print("  "*indent+"+ " + str(len(self.causes)-5) + " more errors", file=file)
			indent -= 1
	def causesContainPartErrors(self):
		return any([cause.containsPartErrors() for cause in self.causes])
	def containsPartErrors(self):
		return self.is_part or self.causesContainPartErrors()

ERROR_STACK = [[]]

def makeErrorMessage(ref, tokens, start, length):
	line1 = ""
	line2 = ""
	line3 = ""
	for i, token in enumerate(tokens):
		inside = start < i < start+length
		char = "~" if inside else " "
		if i != 0:
			line1 += " "
			line2 += char
			if i <= start:
				line3 += " "
		line1 += token.token
		if i == start:
			line2 += "^" + "~" * (len(token.token)-1)
		else:
			line2 += char * len(token.token)
		if i < start:
			line3 += " " * len(token.token)
	return line1 + "\n" + line2 + "\n" + line3

class Pattern:
	def __init__(self, name, words, output):
		self.name = name
		self.words = words
		self.output = output
	def __repr__(self):
		return "Pattern(" + repr(self.name) + ", " + repr(self.words) + ", " + repr(self.output) + ")"
	def toCode(self):
		return " ".join([w.toCode() for w in self.words])# + " -> " + repr(self.output)
	def allowsEmptyContent(self):
		return False
	def match(self, grammar, tokens, bits, i=0, j=0, g=None):
		global indent
		groups = g or {w: [0, []] for w in self.words if isinstance(w, PatternRef)}
		ans = []
		while i <= len(tokens) and j <= len(self.words):
			token = tokens[i] if i < len(tokens) else Token("<END>", [])
			word = self.words[j] if j < len(self.words) else Token("<END>", [])
			if isinstance(word, PatternRef):
				ans += self.match(grammar, tokens, bits, i, j+1, {w: [groups[w][0], groups[w][1].copy()] for w in groups})
				if len(groups[word][1]) == 0:
					groups[word][0] = i
				groups[word][1].append(token)
				i += 1
			else:
				if token.token.lower() == word.token.lower() or token.containsMatch([(wbf, bits if "$" in wbits else wbits) for wbf, wbits in word.alternatives]):
					if debug_level >= 3:
						print(" "*indent+"match:", token, word, bits)
					i += 1
					j += 1
				else:
					if debug_level >= 3:
						print(" "*indent+"no match:", token, word, bits)
					return ans
		if j < len(self.words) or i < len(tokens):
			if debug_level >= 3:
				print(" "*indent+"remainder:", i, tokens, j, self.words)
			return ans
		for w in groups:
			if not groups[w][1] and not grammar.allowsEmptyContent(w.name):
				if debug_level >= 3:
					print(" "*indent+"empty group:", w, self.words, tokens)
				return ans
		
		if debug_level >= 2:
			print(" "*indent, end="")
			for w in self.words:
				print(w.name+"=["+" ".join([str(w2) for w2 in groups[w][1]])+"]" if isinstance(w, PatternRef) else str(w), end=" ")
			print()
			indent += 1
		
		args = []
		for i, w in enumerate([w for w in self.words if isinstance(w, PatternRef)]):
			ERROR_STACK.append([])
			
			match = grammar.matchAll(groups[w][1], w.name, (w.bits-{"$"})|(bits if "$" in w.bits else set()))
			
			errors = ERROR_STACK.pop()
			if len(match) == 0:
				error = ParsingError(
					grammar,
					[w],
					makeErrorMessage(w, tokens, groups[w][0], len(groups[w][1])),
					len(groups[w][1]) != len(tokens),
					errors
				)
				ERROR_STACK[-1].append(error)
			
			args.append(match)
			
			if len(match) == 0:
				break
		ans = ans+[self.output.eval(c) for c in itertools.product(*args)]
		
		if debug_level >= 2:
			indent -= 1
			print(" "*indent, end="")
			print("->", ans)
		
		return ans
